@@ -1,16 +1,25 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-module Network.Wai.Handler.WarpTLS
-    ( TLSSettings
+
+-- | HTTP over SSL/TLS support for Warp via the TLS package.
+
+module Network.Wai.Handler.WarpTLS (
+    -- * Settings
+      TLSSettings
     , certFile
     , keyFile
     , onInsecure
-    , OnInsecure (..)
+    , tlsLogging
+    , defaultTlsSettings
     , tlsSettings
+    , OnInsecure (..)
+    -- * Runner
     , runTLS
     , runTLSSocket
+    -- * Exception
     , WarpTLSException (..)
     ) where
 
@@ -22,7 +31,7 @@ import qualified Data.ByteString.Lazy as L
 import Data.Conduit.Binary (sourceFileRange)
 import qualified Data.Conduit as C
 import qualified Data.Conduit.List as CL
-import Control.Exception (bracket)
+import Control.Exception (bracket, finally)
 import qualified Network.TLS.Extra as TLSExtra
 import qualified Data.Certificate.X509 as X509
 import qualified Data.ByteString as B
@@ -34,11 +43,13 @@ import qualified Data.PEM as PEM
 import Data.Conduit.Network (sourceSocket, sinkSocket, acceptSafe)
 import Data.Maybe (fromMaybe)
 import qualified Data.IORef as I
-import Control.Monad (unless)
-import Crypto.Random.API (getSystemRandomGen)
+import Crypto.Random.AESCtr (makeSystem)
 import Control.Exception (Exception, throwIO)
 import Data.Typeable (Typeable)
 import qualified Data.Conduit.Binary as CB
+#if MIN_VERSION_tls(1, 1, 3)
+import qualified Crypto.Random.AESCtr
+#endif
 
 data TLSSettings = TLSSettings
     { certFile :: FilePath
@@ -58,19 +69,30 @@ data TLSSettings = TLSSettings
       -- Since 1.4.0
     }
 
+-- | An action when a plain HTTP comes to HTTP over TLS/SSL port.
 data OnInsecure = DenyInsecure L.ByteString
                 | AllowInsecure
 
+-- | A smart constructor for 'TLSSettings'.
 tlsSettings :: FilePath -- ^ Certificate file
-            -> FilePath -- ^ key file
+            -> FilePath -- ^ Key file
             -> TLSSettings
-tlsSettings cert key = TLSSettings
+tlsSettings cert key = defaultTlsSettings
     { certFile = cert
     , keyFile = key
+    }
+
+-- | Default 'TLSSettings'. Use this to create 'TLSSettings' with the field record name.
+defaultTlsSettings :: TLSSettings
+defaultTlsSettings = TLSSettings
+    { certFile = "certificate.pem"
+    , keyFile = "key.pem"
     , onInsecure = DenyInsecure "This server only accepts secure HTTPS connections."
     , tlsLogging = TLS.defaultLogging
     }
 
+-- | Running 'Application' with 'TLSSettings' and 'Settings' using
+--   specified 'Socket'.
 runTLSSocket :: TLSSettings -> Settings -> Socket -> Application -> IO ()
 runTLSSocket TLSSettings {..} set sock app = do
     certs   <- readCertificates certFile
@@ -99,11 +121,15 @@ runTLSSocket TLSSettings {..} set sock app = do
                     return bs
             if maybe False ((== 0x16) . fst) (firstBS >>= B.uncons)
                 then do
-                    gen <- getSystemRandomGen
+#if MIN_VERSION_tls(1, 1, 3)
+                    gen <- Crypto.Random.AESCtr.makeSystem
+#else
+                    gen <- makeSystem
+#endif
                     ctx <- TLS.contextNew
                         TLS.Backend
                             { TLS.backendFlush = return ()
-                            , TLS.backendClose = return ()
+                            , TLS.backendClose = sClose s
                             , TLS.backendSend = \bs -> C.yield bs C.$$ toClient
                             , TLS.backendRecv = getNext . fmap (B.concat . L.toChunks) . CB.take
                             }
@@ -116,9 +142,9 @@ runTLSSocket TLSSettings {..} set sock app = do
                             , connSendFile = \fp offset len _th headers _cleaner -> do
                                 TLS.sendData ctx $ L.fromChunks headers
                                 C.runResourceT $ sourceFileRange fp (Just offset) (Just len) C.$$ CL.mapM_ (TLS.sendData ctx . L.fromChunks . return)
-                            , connClose = do
-                                TLS.bye ctx
-                                sClose s
+                            , connClose =
+                                TLS.bye ctx `finally`
+                                TLS.contextClose ctx
                             , connRecv = TLS.recvData ctx
                             }
                     return conn
@@ -142,6 +168,7 @@ data WarpTLSException = InsecureConnectionDenied
     deriving (Show, Typeable)
 instance Exception WarpTLSException
 
+-- | Running 'Application' with 'TLSSettings' and 'Settings'.
 runTLS :: TLSSettings -> Settings -> Application -> IO ()
 runTLS tset set app =
     bracket
